
/*
 *  @(#) obuffer.c 1.6, last edit: 2/21/94 18:10:13
 *  @(#) Copyright (C) 1993, 1994 Tobias Bading (bading@cs.tu-berlin.de)
 *  @(#) Berlin University of Technology
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 /*  Modifications to make maplay work on SPARC1PX systems
     with AMD audio hardware written by
     James Boucher Boston University Multimedia Communications Lab
     email: jboucher@flash.bu.edu      */

/*   ULaw conversion code taken from the Sox Utilities by
     Jef Poskanzer, Craig Reese, and Joe Campbell   */ 

#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include "obuffer.h"
#include "header.h"
#ifdef SPARC1
#include "ulaw.h"     // only need to include this header for ULaw
#endif                // audio devices like the AMD device on SPARC1PXs

extern "C" int ioctl (int, int ...);


ShortObuffer::ShortObuffer (uint32 number_of_channels)
{
#ifdef DEBUG
  if (!number_of_channels || number_of_channels > MAXCHANNELS)
  {
    cerr << "ShortObuffer: number of channels has to be in [1, " <<  MAXCHANNELS << "] !\n";
    exit (1);
  }
#endif
  channels = number_of_channels;
  for (int i = 0; i < number_of_channels; ++i)
    bufferp[i] = buffer + i;
}


void ShortObuffer::append (uint32 channel, int16 value)
{
#ifdef DEBUG
  if (channel >= channels)
  {
    cerr << "illegal channelnumber in ShortObuffer::append()!\n";
    exit (1);
  }
  if (bufferp[channel] - buffer >= OBUFFERSIZE)
  {
    cerr << "ShortObuffer: buffer overflow!\n";
    exit (1);
  }
#endif
  *bufferp[channel] = value;
  bufferp[channel] += channels;
}


void ShortObuffer::write_buffer (int fd)
{
  int length = (int)((char *)bufferp[0] - (char *)buffer);
  if (write (fd, buffer, length) != length)
    cerr << "Warning: couldn't write all samples\n";
  for (int i = 0; i < channels; ++i)
    bufferp[i] = buffer + i;
}



#ifdef Indigo
IndigoObuffer::IndigoObuffer (uint32 number_of_channels, Header *header)
{
#ifdef DEBUG
  if (!number_of_channels || number_of_channels > MAXCHANNELS)
  {
    cerr << "IndigoObuffer: number of channels has to be in [1, " <<  MAXCHANNELS << "] !\n";
    exit (1);
  }
#endif
  channels = number_of_channels;
  for (int i = 0; i < number_of_channels; ++i)
    bufferp[i] = buffer + i;

  // open an audio port and configure it:
  ALconfig config;
  if (!(config = ALnewconfig ()))
  {
    cerr << "ALnewconfig failed!\n";
    exit (1);
  }
  ALsetwidth (config, AL_SAMPLE_16);
  if (channels == 1)
    ALsetchannels (config, AL_MONO);
  else
    ALsetchannels (config, AL_STEREO);
  if (!(port = ALopenport ("MPEG audio player", "w", config)))
  {
    cerr << "can't allocate an audio port!\n";
    exit (1);
  }

  // set sample rate:
  long pvbuffer[2] = { AL_OUTPUT_RATE, 0 };
  pvbuffer[1] = header->frequency ();
  ALsetparams (AL_DEFAULT_DEVICE, pvbuffer, 2);
  ALfreeconfig (config);
}


IndigoObuffer::~IndigoObuffer (void)
{
  while (ALgetfilled (port) > 0)
    sleep (1);
  ALcloseport (port);
}


void IndigoObuffer::append (uint32 channel, int16 value)
{
#ifdef DEBUG
  if (channel >= channels)
  {
    cerr << "illegal channelnumber in IndigoObuffer::append()!\n";
    exit (1);
  }
  if (bufferp[channel] - buffer >= OBUFFERSIZE)
  {
    cerr << "IndigoObuffer: buffer overflow!\n";
    exit (1);
  }
#endif
  *bufferp[channel] = value;
  bufferp[channel] += channels;
}


void IndigoObuffer::write_buffer (int)
{
  ALwritesamps (port, buffer, (long)(bufferp[0] - buffer));
  for (int i = 0; i < channels; ++i)
    bufferp[i] = buffer + i;
}
#endif // Indigo

#ifndef SPARC1    // If we are not using AMD device
#ifdef SPARC      // so we must be using the SPARC10's dbri device
int SparcObuffer::audio_fd = -1;

SparcObuffer::SparcObuffer (uint32 number_of_channels, Header *header,
			    bool use_speaker, bool use_headphone, bool use_line_out)
{
#ifdef DEBUG
  if (!number_of_channels || number_of_channels > MAXCHANNELS)
  {
    cerr << "SparcObuffer: 0 < number of channels < " << MAXCHANNELS << "!\n";
    exit (1);
  }
#endif
  channels = number_of_channels;
  for (int i = 0; i < number_of_channels; ++i)
    bufferp[i] = buffer + i;

  if (audio_fd < 0)
  {
    cerr << "Internal error, SparcObuffer::audio_fd has to be initialized\n"
	    "by SparcObuffer::class_suitable()!\n";
    exit (1);
  }

  // configure the device:
  audio_info info;
  AUDIO_INITINFO (&info);
  info.output_muted = False;
  info.play.encoding = AUDIO_ENCODING_LINEAR;
  info.play.precision = 16;
  info.play.channels = channels;
  info.play.sample_rate = header->frequency ();
  if (use_speaker)
    info.play.port |= AUDIO_SPEAKER;
  if (use_headphone)
    info.play.port |= AUDIO_HEADPHONE;
  if (use_line_out)
    info.play.port |= AUDIO_LINE_OUT;
  info.play.balance = AUDIO_MID_BALANCE;
  if (ioctl (audio_fd, AUDIO_SETINFO, &info))
  {
    perror ("configuration of /dev/audio failed");
    exit (1);
  }
}


SparcObuffer::~SparcObuffer (void)
{
  sleep (1);
  close (audio_fd);
}


void SparcObuffer::append (uint32 channel, int16 value)
{
#ifdef DEBUG
  if (channel >= channels)
  {
    cerr << "illegal channelnumber in SparcObuffer::append()!\n";
    exit (1);
  }
  if (bufferp[channel] - buffer >= OBUFFERSIZE)
  {
    cerr << "buffer overflow!\n";
    exit (1);
  }
#endif
  *bufferp[channel] = value;
  bufferp[channel] += channels;
}


void SparcObuffer::write_buffer (int)
{
  int length = (int)((char *)bufferp[0] - (char *)buffer);
  if (write (audio_fd, buffer, length) != length)
    cerr << "Warning: couldn't write all samples to /dev/audio\n";
  for (int i = 0; i < channels; ++i)
    bufferp[i] = buffer + i;
}


int SparcObuffer::open_audio_device (void)
{
  int fd;

  if ((fd = open ("/dev/audio", O_WRONLY | O_NDELAY)) < 0)
    if (errno == EBUSY)
    {
      cerr << "Sorry, the audio device is busy!\n";
      exit (1);
    }
    else
    {
      perror ("can't open /dev/audio for writing");
      exit (1);
    }
  return fd;
}


#ifdef Solaris
void SparcObuffer::get_device_type (int fd, audio_device *devtype)
{
  if (ioctl (fd, AUDIO_GETDEV, devtype))
  {
    perror ("ioctl on /dev/audio");
    exit (1);
  }
}
#else
int SparcObuffer::get_device_type (int fd)
{
#ifdef AUDIO_GETDEV
  int devtype;
  if (ioctl (fd, AUDIO_GETDEV, &devtype))
  {
    perror ("ioctl on /dev/audio");
    exit (1);
  }
  return devtype;
#else
  cerr << "Warning: AUDIO_GETDEV ioctl not available!\n";
  return -1;
#endif
}
#endif // !Solaris


bool SparcObuffer::class_suitable (void)
{
  // check for the dbri audio device:
  audio_fd = open_audio_device ();
#ifdef Solaris
  audio_device devtype;
  get_device_type (audio_fd, &devtype);
  if (!strcmp (devtype.name, "SUNW,dbri"))
#else
  if (get_device_type (audio_fd) == AUDIO_DEV_SPEAKERBOX)
#endif
    return True;
  else
    return False;
}
#endif

#else // SPARC1  - This section copies that above except for
      // the changes to the device being used

int SparcObuffer::audio_fd = -1;

// Sparc Output buffer constructor
SparcObuffer::SparcObuffer (uint32 number_of_channels, Header *header,
			    bool use_speaker, bool use_headphone, bool use_line_out)
{
#ifdef DEBUG
  if (!number_of_channels || number_of_channels > MAXCHANNELS)
  {
    cerr << "SparcObuffer: 0 < number of channels < " << MAXCHANNELS << "!\n";
    exit (1);
  }
#endif
  channels = number_of_channels;  // can be either 1 or 2 for input file
  for (int i = 0; i < number_of_channels; ++i)
    bufferp[i] = buffer + i;  // use single interleaved buffer for
                              // 2 channel output, but use two
                              // different pointers into it

/* We need to save a copy of the input data sample rate
   here so the write_buffer() routine can set its internal
   state properly.  The variable samp_freq is visible at file scope */

samp_freq = header->frequency();  
}

// Destructor
SparcObuffer::~SparcObuffer (void)
{
  sleep (1);
}

/*  The append method takes individual channels and
    a 16 bit signed sample and appends the sample
    to the output buffer.  The AMD audio device plays
    out only one channel(left) of 8 bit ULAW data
    sampled at 8KHz.  Therefore, we need to convert
    the 16 bit signed values to ULAW before we
    append them to the buffer. Note that we are
    throwing away the right channel here. This is
    the fastest way to go stereo to mono  */
void SparcObuffer::append (uint32 channel, int16 value)
{
    int sign, exponent, mantissa;
    unsigned char ulawbyte;      // ulaw value written to buffer
#ifdef DEBUG
  if (channel >= channels)
  {
    cerr << "illegal channelnumber in SparcObuffer::append()!\n";
    exit (1);
  }
  if (bufferp[channel] - buffer >= OBUFFERSIZE)
  {
    cerr << "buffer overflow!\n";
    exit (1);
  }
#endif

/* Ignore channel 1 since output is mono */
if(channel == 0)
 {

/* convert value from 16 bit linear signed to 8 bit ULaw */

    /* Get the sample into sign-magnitude. */
    sign = (value >> 8) & 0x80;                /* set aside the sign */
    if ( sign != 0 ) value = -value;          /* get magnitude */
    if ( value > 32635 ) value = 32635;         /* clip the magnitude */

    /* Convert from 16 bit linear to ulaw. */
    value = value + 0x84;
    exponent = exp_lut[( value >> 7 ) & 0xFF];
    mantissa = ( value >> ( exponent + 3 ) ) & 0x0F;
    ulawbyte = ~ ( sign | ( exponent << 4 ) | mantissa );
    if ( ulawbyte == 0 ) ulawbyte = 0x02;       /* optional CCITT trap */
  
  *bufferp[channel] = ulawbyte;
  bufferp[channel] += 1;
  }
}

/* At this point, we have a buffer of 8 bit ulaw values of
   length 1152 for layer II and 384 for layer I. The sample
   frequency is either 32, 44.1, or 48 KHz.  We need a fast
   way to convert from these higher rates to the output
   rate of 8KHz. FAST is the most important factor here
   since the SPARC1PX is only barely powerful enough to
   playback low bitrate streams. Any additional computational
   burden would make the product unusable.  At this time,
   music played at layer II - 32 to 64 Kbps plays out
   fine(AM radio) and layer I- 64 Kbps plays out fine.  Given the
   linited quality of SPARC1PX playback, listening tests
   found little difference between the layer II 32 and 64
   Kbps streams.  Therefore, compress your .au files
   or PCM files to 32 Kbps layer II to save on processing
   overhead.  Note that 32 Kbps forces single channel
   encoding.     */
   
void SparcObuffer::write_buffer (int)
{
 /* The fastest way to convert 48 KHz to 8KHz is to
    start at the head of the buffer and copy out
    every 6th(48/8) sample into a new output buffer.
    This may sound sloppy, but when your input stream
    is already limited to 64 Kbps, it is like you
    have already performed a lowpass filtering on
    the data.  Although we could have averaged
    intermediate samples, I tried it an found little
    or no audible difference.  Given the added
    computational overhead, the averaging was deemed
    unneccessary.  */

  char buffer8K[288];   // true output buffer for 8KHz samples
  float space = 0.0;    // space stores the frequency gap size 
  int length8K = 0;     // the length of the true output buffer

  int length = (int)((char *)bufferp[0] - (char *)buffer);
  if(samp_freq == 48000){
    space = 6.0;         // 48000/8000 = 6 - take every 6th sample
    length8K = length/6;    // The true output buffer is 1/6 as big
  } else if(samp_freq == 44100){
    space = 5.5;         // 44100/8000 = 5.5125 rounded to 5.5
    length8K = (int)(length/5.5);
  } else if(samp_freq == 32000){
    space = 4.0;         // 32000/8000 = 4 
    length8K = (length>>2);
  }else{
    cerr<<"Bad sample_frequency\n";
    exit(1);
  }
// strip out only the samples we want
for(int x=0;x<length8K;x++){
buffer8K[x] = buffer[(int)(x*space)];
}
/* To output to the AMD device you need only write
   the data into /dev/audio. If you are trying
   to output too high a bitrate, you will get
   a breaking up sound from the output */
  if (write (audio_fd, buffer8K, length8K) != length8K)
    cerr << "Warning: couldn't write all samples to /dev/audio\n";
  for (int i = 0; i < channels; ++i)
    bufferp[i] = buffer + i;
}


int SparcObuffer::open_audio_device (void)
{
  int fd;

  if ((fd = open ("/dev/audio", O_WRONLY | O_NDELAY)) < 0)
    if (errno == EBUSY)
    {
      cerr << "Sorry, the audio device is busy!\n";
      exit (1);
    }
    else
    {
      perror ("can't open /dev/audio for writing");
      exit (1);
    }
  return fd;
}

bool SparcObuffer::class_suitable (void)
{
  // check for the AMD audio device:

    audio_fd = open_audio_device();
    return True;
}
#endif

